#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""

Input & Output

    fclean, write_fits, read_fits, write_hdf5, read_hdf5,
    write_ascii, read_ascii, write_csv, read_csv

"""

import sys, logging
logging.disable(sys.maxsize)

import subprocess as SP
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
import h5py as H5
import csv

global fitsext, h5ext, ascext, csvext, savext
fitsext = '.fits'
h5ext = '.h5'
ascext = '.txt'
csvext = '.csv'
savext = '.sav'

def fclean(file, *alert):
    '''
    Clean folder/files
    '''
    SP.call('rm -rf '+file, shell=True)
    for text in alert:
        print(text)

def write_fits(file, header, data, wave=None, wmod=0, **hdrl):
    '''
    Write fits file

    ------ INPUT ------
    file                output FITS filename
    header              header of primary HDU
    data                data in primary HDU
    wave                data in table 1 (ndarray. Default: None)
    wmod                wave table format (0 - Image; 1 - BinTable. Default: 0)
    ------ OUTPUT ------
    '''
    for key, value in hdrl.items():
        header[key] = value
    primary_hdu = fits.PrimaryHDU(header=header, data=data)
    hdul = fits.HDUList(primary_hdu)
    
    ## Add table
    if wave is not None:
        ## Convert wave format
        if isinstance(wave, fits.fitsrec.FITS_rec):
            if wmod==0:
                wave = wave[0][0][:,0]
        else:
            Nw = len(wave)
            if wmod==1:
                wave = np.array(wave).reshape((Nw,1))
                col = fits.Column(array=[wave], format=str(Nw)+'E',
                                  name='WAVE-TAB', unit='um', dim='(1,{})'.format(Nw))
                tab = fits.BinTableHDU.from_columns([col], name='WCS-TAB ')
                wave = tab.data
        ## Create table
        if wmod==0:
            hdu = fits.ImageHDU(data=wave, name='WAVE-TAB')
        elif wmod==1:
            hdu = fits.BinTableHDU(data=wave, name='WCS-TAB ')

        hdul.append(hdu)

    hdul.writeto(file+fitsext, overwrite=True)
    
def read_fits(file, file_unc=None, wmod=0):
    '''
    Read fits file (auto detect dim)

    ------ INPUT ------
    file                input FITS filename
    file_unc            input uncertainty file
    wmod                output wave mode (Default: 0)
                          0 - 1darray; 
                          1 - FITS_rec.
    ------ OUTPUT ------
    ds                  output dataset
      header              header of primary HDU
      data                data in primary HDU
      header_w            header of W-TAB
      wave                data in table 1 (None if does not exist)
      unc                 uncertainty array
    '''
    ## Initialize output object
    ds = type('', (), {})()
    ds.header_w = None
    ds.wave = None
    ds.unc = None
    
    ## Read header & data
    with fits.open(file+fitsext) as hdul:
        ds.HDUL = hdul
        hdr = hdul[0].header
        ds.data = hdul[0].data
        ds.header = hdr
        try:
            ds.wcs = WCS(hdr)
        except:
            ds.wcs = WCS(None)
            # print('Not supported WCS (probably with 3D SIP distortion), use astrom.fixwcs')

        ## Read wavelength
        if len(hdul)==2:
            ds.header_w = hdul[1].header
            wave = hdul[1].data

            if isinstance(hdul[1], fits.BinTableHDU):
                if wmod==0:
                    wave = wave[0][0][:,0] ## Convert FITS_rec to 1darray
            elif isinstance(hdul[1], fits.ImageHDU):
                Nw = len(wave)
                if wmod==1:
                    wave = np.array(wave).reshape((Nw,1))
                    col = fits.Column(array=[wave], format=str(Nw)+'E',
                                      name='WAVE-TAB', unit='um', dim='(1,{})'.format(Nw))
                    tab = fits.BinTableHDU.from_columns([col], name='WCS-TAB ')
                    wave = tab.data
            
            ds.wave = wave
    
    if file_unc is not None:
        ## Read uncertainty data
        with fits.open(file_unc+fitsext) as hdul:
            ds.unc = hdul[0].data

    return ds

def write_hdf5(file, name, data, group='/',
               ind1=None, ind2=None, ind3=None, ind4=None,
               ind5=None, ind6=None, ind7=None,
               append=False, verbose=False):
    '''
    Write dataset into a h5 file (a single name/data_array per time, dim < 7)
    Inspired by SwING inout library

    ------ INPUT ------
    file                output h5 filename
    name                name of the dataset (len <= 80)
    data                dataset (dim < 7)
    group               name of the group (Default: '/')
    indx                array index ([idimx_inf,idimx_sup] or idimx if data is scalar, Default: None)
    append              True: if not overwrite (Default: False)
    verbose             courtesy notification (Default: False)
    ------ OUTPUT ------
    '''
    ## Preliminaries
    ##---------------
    if (append):
        hf = H5.File(file+h5ext, 'a')
    else:
        hf = H5.File(file+h5ext, 'w')
    
    h5 = hf.require_group(group) # Default: group = '/'

    ## Check if it is a subarray
    subarr = ( (ind1 is not None) or (ind2 is not None) or (ind3 is not None) or \
               (ind4 is not None) or (ind5 is not None) or (ind6 is not None) )

    ## Convert all data to array format
    darr = np.array(data)
    sharr = darr.shape
    if darr.dtype.kind == 'U':
        ## create_dataset does not support lists of UTF-8 yet
        ## "these strings are supposed to store only ASCII-encoded text"
        ## See http://docs.h5py.org/en/stable/strings.html
        asciiList = [n.encode('ascii','ignore') for n in darr.flatten()]
        darr = np.array(asciiList).reshape(sharr)

    ## Write dataset (dset)
    ##----------------------
    
    ## Case 0: check if the dataset already exists
    try:
        dset = h5[name]
        hasdset = True
    except:
        hasdset = False
    if (hasdset):
        createdset = False
        ## No indx input
        if (not subarr):
            del h5[name]
            createdset = True
    else:
        createdset = True
        if (subarr):
            h5.close()
            raise ValueError('Dataset does not exist')

    ## Case 1: no dataset OR no subarr
    if (createdset):
        dset = h5.create_dataset(name, data=darr)

    ## Case 2: sub-array filling
    if (subarr):
        ## Dimension of the array
        Ndim = len(sharr)

        ## Indices indx[idimx_inf,idimx_sup]
        if (Ndim > 0):
            if ind1 is None:
                ind1 = [0,sharr[0]]
            elif (np.size(ind1) == 1):
                ind1 = [ind1, ind1+1]
        if (Ndim > 1):
            if ind2 is None:
                ind2 = [0,sharr[1]]
            elif (np.size(ind2) == 1):
                ind2 = [ind2, ind2+1]
        if (Ndim > 2):
            if ind3 is None:
                ind3 = [0,sharr[2]]
            elif (np.size(ind3) == 1):
                ind3 = [ind3, ind3+1]
        if (Ndim > 3):
            if ind4 is None:
                ind4 = [0,sharr[3]]
            elif (np.size(ind4) == 1):
                ind4 = [ind4, ind4+1]
        if (Ndim > 4):
            if ind5 is None:
                ind5 = [0,sharr[4]]
            elif (np.size(ind5) == 1):
                ind5 = [ind5, ind5+1]
        if (Ndim > 5):
            if ind6 is None:
                ind6 = [0,sharr[5]]
            elif (np.size(ind6) == 1):
                ind6 = [ind6, ind6+1]
        if (Ndim > 6):
            if ind7 is None:
                ind7 = [0,sharr[6]]
            elif (np.size(ind7) == 1):
                ind7 = [ind7, ind7+1]

        ## Write the sub-array if indx are set
        if Ndim == 0:
            dset = arr[0]
        elif Ndim == 1:
            dset[ind1[0]:ind1[1]] = np.reshape( darr, (ind1[1]-ind1[0],) )
        elif Ndim == 2:
            dset[ind1[0]:ind1[1],ind2[0]:ind2[1]] = \
                np.reshape( darr, (ind1[1]-ind1[0],ind2[1]-ind2[0]) )
        elif Ndim == 3:
            dset[ind1[0]:ind1[1],ind2[0]:ind2[1],ind3[0]:ind3[1]] = \
                np.reshape( darr, (ind1[1]-ind1[0],ind2[1]-ind2[0],ind3[0]-ind3[1]) )
        elif Ndim == 4:
            dset[ind1[0]:ind1[1],ind2[0]:ind2[1],ind3[0]:ind3[1],ind4[0]:ind4[1]] = \
                np.reshape( darr, (ind1[1]-ind1[0],ind2[1]-ind2[0], \
                                   ind3[0]-ind3[1],ind4[0]-ind4[1]) )
        elif Ndim == 5:
            dset[ind1[0]:ind1[1],ind2[0]:ind2[1],ind3[0]:ind3[1], \
                 ind4[0]:ind4[1],ind5[0]:ind5[1]] = \
                np.reshape( darr, (ind1[1]-ind1[0],ind2[1]-ind2[0], \
                                   ind3[0]-ind3[1],ind4[0]-ind4[1], \
                                   ind5[0]-ind5[1]) )
        elif Ndim == 6:
            dset[ind1[0]:ind1[1],ind2[0]:ind2[1],ind3[0]:ind3[1], \
                 ind4[0]:ind4[1],ind5[0]:ind5[1],ind6[0]:ind6[1]] = \
                np.reshape( darr, (ind1[1]-ind1[0],ind2[1]-ind2[0], \
                                   ind3[0]-ind3[1],ind4[0]-ind4[1], \
                                   ind5[0]-ind5[1],ind6[0]-ind6[1]) )
        elif Ndim == 7:
            dset[ind1[0]:ind1[1],ind2[0]:ind2[1],ind3[0]:ind3[1], \
                 ind4[0]:ind4[1],ind5[0]:ind5[1],ind6[0]:ind6[1],ind7[0]:ind7[1]] = \
                np.reshape( darr, (ind1[1]-ind1[0],ind2[1]-ind2[0], \
                                   ind3[0]-ind3[1],ind4[0]-ind4[1], \
                                   ind5[0]-ind5[1],ind6[0]-ind6[1],ind7[0]-ind7[1]) )

    hf.flush()
    hf.close()
    
    ## Courtesy notification
    ##-----------------------
    if (verbose):
        if (append):
            if (subarr):
                print('[write_hdf5] Dataset {}/{} in the file:'.format(group,name))
                print('    ', file+h5ext, ' has been modified.')
            else:
                print('[write_hdf5] Dataset {}/{} has been added in the file:'.format(group,name))
                print('    ', file+h5ext, '.')
        else:
            print('[write_hdf5] Dataset {}/{} has been written in the new file:'.format(group,name))
            print('    ', file+h5ext, '.')

def read_hdf5(file, name, group='/',
              ind1=None, ind2=None, ind3=None, ind4=None,
              ind5=None, ind6=None, ind7=None):
    '''
    Read h5 file (a single name/data_array per time, dim < 7)
    Inspired by SwING inout library

    ------ INPUT ------
    file                input h5 filename
    name                name of the dataset (len <= 80)
    group               name of the group (Default: '/')
    indx                array index ([idimx_inf,idimx_sup] or idimx if data is scalar, Default: None)
    ------ OUTPUT ------
    dset                dataset
    '''
    ## Preliminaries
    ##---------------
    hf = H5.File(file+h5ext, 'r')
    h5 = hf.require_group(group) # Default: group = '/'

    Ndim = h5[name].ndim
    shapIN = h5[name].shape

    ## Read dataset (dset)
    ##----------------------
    
    ## Indices indx[idimx_inf,idimx_sup]
    if (Ndim > 0):
        if ind1 is None:
            ind1 = [0,shapIN[0]]
        elif (np.size(ind1) == 1):
            ind1 = [ind1, ind1+1]
    if (Ndim > 1):
        if ind2 is None:
            ind2 = [0,shapIN[1]]
        elif (np.size(ind2) == 1):
            ind2 = [ind2, ind2+1]
    if (Ndim > 2):
        if ind3 is None:
            ind3 = [0,shapIN[2]]
        elif (np.size(ind3) == 1):
            ind3 = [ind3, ind3+1]
    if (Ndim > 3):
        if ind4 is None:
            ind4 = [0,shapIN[3]]
        elif (np.size(ind4) == 1):
            ind4 = [ind4, ind4+1]
    if (Ndim > 4):
        if ind5 is None:
            ind5 = [0,shapIN[4]]
        elif (np.size(ind5) == 1):
            ind5 = [ind5, ind5+1]
    if (Ndim > 5):
        if ind6 is None:
            ind6 = [0,shapIN[5]]
        elif (np.size(ind6) == 1):
            ind6 = [ind6, ind6+1]
    if (Ndim > 6):
        if ind7 is None:
            ind7 = [0,shapIN[6]]
        elif (np.size(ind7) == 1):
            ind7 = [ind7, ind7+1]
        
    # Read the array or the sub-array if some indx are set
    if Ndim==0:
        dset = h5[name]
        if (str(dset.dtype).startswith('|S')):
            dset = np.array(dset, dtype='unicode')
    elif Ndim==1:
        dset = h5[name][ind1[0]:ind1[1]]
        if (str(dset.dtype).startswith('|S')):
            dset = np.array(dset, dtype='unicode')
    elif Ndim==2:
        dset = h5[name][ind1[0]:ind1[1],ind2[0]:ind2[1]]
        if (str(dset.dtype).startswith('|S')):
            dset = np.array(dset, dtype='unicode')
    elif Ndim==3:
        dset = h5[name][ind1[0]:ind1[1],ind2[0]:ind2[1],ind3[0]:ind3[1]]
        if (str(dset.dtype).startswith('|S')):
            dset = np.array(dset, dtype='unicode')
    elif Ndim==4:
        dset = h5[name][ind1[0]:ind1[1],ind2[0]:ind2[1],ind3[0]:ind3[1], \
                        ind4[0]:ind4[1]]
        if (str(dset.dtype).startswith('|S')):
            dset = np.array(dset, dtype='unicode')
    elif Ndim==5:
        dset = h5[name][ind1[0]:ind1[1],ind2[0]:ind2[1],ind3[0]:ind3[1], \
                        ind4[0]:ind4[1],ind5[0]:ind5[1]]
        if (str(dset.dtype).startswith('|S')):
            dset = np.array(dset, dtype='unicode')
    elif Ndim==6:
        dset = h5[name][ind1[0]:ind1[1],ind2[0]:ind2[1],ind3[0]:ind3[1], \
                        ind4[0]:ind4[1],ind5[0]:ind5[1],ind6[0]:ind6[1]]
        if (str(dset.dtype).startswith('|S')):
            dset = np.array(dset, dtype='unicode')
    elif Ndim==7:
        dset = h5[name][ind1[0]:ind1[1],ind2[0]:ind2[1],ind3[0]:ind3[1], \
                        ind4[0]:ind4[1],ind5[0]:ind5[1],ind6[0]:ind6[1], \
                        ind7[0]:ind7[1]]
        if (str(dset.dtype).startswith('|S')):
            dset = np.array(dset, dtype='unicode')

    hf.close()

    return dset

def write_ascii(file, header=None, dset=None, trans=False,
                ascext=ascext, append=False, comment=None):
    '''
    Write ASCII file
    Supported format: commented_header
    See also: astropy.io.ascii.write, numpy.savetxt
    
    ------ INPUT ------
    file                input ASCII filename
    header              data header
    dset                dataset (1darray[nrow], 2darray[nrow,ncol] or list)
    trans               transpose dset (Default: False)
    ascext              ASCII file suffix (Default: '.txt')
    append              True: if not overwrite (Default: False)
    comment             single line comment on top
    ------ OUTPUT ------
    '''
    if append==True:
        mod = 'a'
    else:
        mod = 'w'

    with open(file+ascext, mod) as f:
        pass
        ## Comment on the top
        ##--------------------
        if comment is not None:
            f.write('## '+comment+'\n')
        
        ## Number of columns/rows
        ncol = 0
        nrow = 0
        ## Dataset
        ##---------
        rows = ''
        if dset is not None:
            dset = np.array(dset)
            ## Transpose dset
            if trans==True:
                Ny, Nx = dset.shape
                darr = []
                for x in range(Nx):
                    for y in range(Ny):
                        darr.append(dset[y,x])
                darr = np.array(darr).reshape((Nx,Ny))
            else:
                darr = dset
            ## Update number of columns/rows
            Ndim = darr.ndim
            if Ndim==1:
                ncol = 1
                nrow = len(darr)
            else:
                nrow, ncol = darr.shape
            ## Convert dataset array to string
            if Ndim==1 or Ndim==2:
                if nrow<500:
                    rows += str(darr).replace('[','').replace(']','')+'\n'
                else:
                    ## Python 500 row limit
                    for i in range(nrow//500+1):
                        rows += str(darr[500*i:500*(i+1)]).replace('[','').replace(']','')+'\n'
            else:
                raise ValueError(
                    'Expected 1D or 2D array, got {}D array instead!'.format(Ndim))
        ## Header
        ##--------
        hrow = ''
        if header is not None:
            if isinstance(header, str):
                header = [header]
            else:
                header = list(header)
                
            for i, h in enumerate(header):
                if i==0:
                    hrow += h
                else:
                    hrow += ' '+h
            hrow += '\n'
            ## Update number of columns/rows
            ncol = len(header)
            nrow += 1
        ## Column width control
        ##----------------------
        ## Table in a single string
        if nrow==0:
            fwrows = ''
        else:
            cells = hrow.split()
            cells.extend(rows.split())
            cells = np.array(cells).reshape((-1,ncol))
            col_width = []
            for x in range(ncol):
               col_width.append(len(max(cells[:,x], key=len)))
            ## Table with fixed column widths in a single string
            if hrow=='':
                fwrows = '' # no header
            else:
                fwrows = '# '
            for y in range(nrow):
                for x in range(ncol):
                    c = cells[y,x]
                    ## '{message:{fill}{align}{width}}'.format(
                    ##     message=c, fill=' ', align='>', width=col_width+1)
                    if y==0 and x==0 and header is not None:
                        ## Due to hashtag
                        fwrows += ''.ljust(col_width[x]-len(c))+c
                    else:
                        fwrows += ''.ljust(col_width[x]+2-len(c))+c
                    if x==ncol-1:
                        fwrows += '\n'

        f.write(fwrows)

def read_ascii(file, ascext=ascext, dtype=str, start_header=-1):
    '''
    Read ASCII file
    Supported format: commented_header
    See also: astropy.io.ascii.read, numpy.genfromtxt

    ------ INPUT ------
    file                input ASCII filename
    ascext              ASCII file suffix (Default: '.txt')
    dtype               data type (Default: 'str')
    start_header        line number of the header line (Default: -1)
    ------ OUTPUT ------
    dset                dataset (dict)
    '''
    with open(file+ascext, 'r') as f:
        ## f.read() -> str | f.readlines() -> list
        header = None
        darr = []
        for i, line in enumerate(f.readlines()):
            line = line.strip()
            # print(line)
            if line[0]=='#':
                ## By default, header is taken from the last commented line before data
                if isinstance(start_header, int)==True:
                    if start_header==-1:
                        header = line.split()[1:]
                    elif start_header>0 and i==start_header-1:
                        header = line.split()[1:]
            else:
                line = list(map(dtype, line.split()))
                row = []
                for cell in line:
                    row.append(cell)
                darr.append(row)

    darr = np.array(darr)

    ## Number of columns/rows
    nrow, ncol = darr.shape

    ## Default header = ['col1','col2',...]
    header_default = []
    for x in range(ncol):
        header_default.append('col'+str(x+1))

    ## Dict out
    dset = {}
    for x, h in enumerate(header_default): 
        dset[h] = darr[:,x]
        if header is not None:
            dset[header[x]] = darr[:,x]

    return dset

def write_csv(file, header, dset, trans=False, append=False):
    '''
    Read fits file
    See also: astropy.io.ascii.write

    ------ INPUT ------
    file                output csv filename
    header              data labels in list('label1', 'label2', ...)
    dset                dataset (1darray[nrow], 2darray[nrow,ncol] or list)
    trans               transpose dset (Default: False)
    append              True: if not overwrite (Default: False)
    ------ OUTPUT ------
    '''
    if append==True:
        mod = 'a'
    else:
        mod = 'w'

    with open(file+csvext, mod, newline='') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=header)

        ## Header
        ##--------
        writer.writeheader()

        ## Dataset
        ##---------
        dset = np.array(dset)
        ## Transpose dset
        if trans==True:
            Ny, Nx = dset.shape
            darr = []
            for x in range(Nx):
                for y in range(Ny):
                    darr.append(dset[y,x])
            darr = np.array(darr).reshape((Nx,Ny))
        else:
            darr = dset
        ## Number of columns/rows
        Ndim = darr.ndim
        if Ndim==1:
            ncol = 1
            nrow = darr.shape[0]
        else:
            nrow, ncol = darr.shape
        ## Write
        for y in range(nrow):
            ## Init dict
            rows = {h: [] for h in header}
            ## Write dict
            for x in range(ncol):
                rows[header[x]] = darr[y,x]
            ## Write csv row
            writer.writerow(rows)
            
def read_csv(file, *header):
    '''
    Read csv file

    ------ INPUT ------
    file                input csv filename
    header              labels of data to read
    ------ OUTPUT ------
    dset                dataset (dict)
    '''
    with open(file+csvext, 'r', newline='') as csvfile:
        dset = {}
        for h in header:
            csvfile.seek(0) # reset pointer
            reader = csv.DictReader(csvfile) # pointer at end
            col = []
            for row in reader:
                col.append(row[h])
            data = np.array(col)
            dset[h] = data

    return dset



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""

Calibration

    intercalib:
        read_filter, synthetic_photometry, correct_spec
    photometry_profile

"""

import os
import math
import numpy as np
# from astropy.io import ascii
from matplotlib.ticker import ScalarFormatter, NullFormatter
import subprocess as SP
import warnings
DEVNULL = open(os.devnull, 'w')

## Local
from arrays import listize, pix2sup, sup2pix
from inout import (ascext, fitsext, h5ext,
                   read_fits, write_fits,
                   read_hdf5, write_hdf5#, read_ascii
)
from astrom import fixwcs
from plots import pplot

## Path of current file
croot = os.path.dirname(os.path.abspath(__file__))

##-----------------------------------------------
##
##            <intercalib> based tools
##
##-----------------------------------------------

class intercalib:
    '''
    Intercalibration

    ------ INPUT ------
    filIN               target FITS file (Default: None)
    '''
    def __init__(self, filIN=None):
        
        ## INPUTS
        self.filIN = filIN

        if filIN is not None:
            # self.hdr = fixwcs(filIN+fitsext).header
            w = fixwcs(filIN+fitsext).wcs
            ds = read_fits(filIN)
            self.hdr = ds.header
            self.im = ds.data
            self.wvl = ds.wave

    def read_filter(self, filt, w_spec=None):
        '''
        Return center wavelength of the filters
        The input offset corresponds to the integrated broad band value (bboff).
        Suppose a flat spectral offset to correct the spectral value (specoff).
        Return specoff/bboff

        ------ INPUT ------
        filt                photometry names (string, tuple or list)
        w_spec              wavelengths (Default: None - via filIN)
        ------ OUTPUT ------
        self
          wcen                center wavelength
          specoff_ov_bboff    spectral/broad band offset ratio
        '''
        ## Convert all format phot names to list
        filt = listize(filt)

        if self.filIN is not None:
            w_spec = self.wvl
        for phot in filt:
            w_grid = read_hdf5(croot+'/lib/data/filt_'+phot,
                               'Filter wavelength (microns)')
            if w_spec[0]>w_grid[0] or w_spec[-1]<w_grid[-1]:
                warnings.warn('Synthetic photometry of {} can be underestimated' \
                              'due to uncovered wavelengths. '.format(phot))
        ## Insert 2 wvl (0.01 um & w_spec[0]-0.01 um) with 0 value
        wave = np.insert(w_spec, 0, (.01, w_spec[0]-.01))
        Fnu_uni = np.ones(len(wave))
        Fnu_uni[:2] = 0
        flux = Fnu_uni[:,np.newaxis,np.newaxis]

        ## Write input.h5
        ##----------------
        fortIN = os.getcwd()+'/synthetic_photometry_input'
        
        write_hdf5(fortIN, 'Filter label', filt)
        write_hdf5(fortIN, 'Wavelength (microns)', wave, append=True)
        write_hdf5(fortIN, 'Flux (x.Hz-1)', flux, append=True)
        write_hdf5(fortIN, '(docalib,dophot)', [1,1], append=True)

        ## Call the Fortran lib
        ##----------------------
        SP.call('synthetic_photometry', shell=True)

        ## Read output.h5
        ##----------------
        fortOUT = os.getcwd()+'/synthetic_photometry_output'

        self.wcen = read_hdf5(fortOUT, 'Central wavelength (microns)')
        Fnu_filt = read_hdf5(fortOUT, 'Flux (x.Hz-1)')[:,0,0]
        self.specoff_ov_bboff = 1. / Fnu_filt

        ## Clean temperary h5 files
        ##--------------------------
        SP.call('rm -rf '+fortIN+h5ext, shell=True, cwd=os.getcwd())
        SP.call('rm -rf '+fortOUT+h5ext, shell=True, cwd=os.getcwd())

    def synthetic_photometry(self, filt, w_spec=None, Fnu_spec=None,
                             xscale=1, yscale=1, 
                             extrapoff=True, verbose=False):
        '''
        External Fortran library (SwING) needed

        ------ INPUT ------
        filt                photometry names (string, tuple or list)
        w_spec              wavelengths (Default: None - via filIN)
        Fnu_spec            spectra (Default: None - via filIN)
        extrapoff           set zeros for uncovered wave grid (Default: True)
        verbose             keep tmp files (Default: False)
        ------ OUTPUT ------
        ds                  output dataset
          wcen              center wavelength
          Fnu_filt          synthetic photometry
          smat              standard deviation matrices
        '''
        ds = type('', (), {})()

        ## Convert all format phot names to list
        filt = listize(filt)

        ## Input is a FITS file
        if self.filIN is not None:
            w_spec = self.wvl
            Fnu_spec = self.im

        w_spec = np.array(w_spec)
        Fnu_spec = np.array(Fnu_spec)
        if len(Fnu_spec.shape)==1:
            Ndim = 1
            Fnu_spec = Fnu_spec[:,np.newaxis,np.newaxis]
        else:
            Ndim = 3

        ## Super pixels
        Nw, Ny, Nx = Fnu_spec.shape
        Nxs = math.ceil(Nx/xscale)
        spec_supx = np.zeros((Nw,Ny,Nxs))
        for xs in range(Nxs):
            x = sup2pix(xs, xscale, Npix=Nx, origin=0)
            spec_supx[:,:,xs] += np.nanmean(Fnu_spec[:,:,x[0]:x[-1]+1],axis=2)
        Nys = math.ceil(Ny/yscale)
        sup_spec = np.zeros((Nw,Nys,Nxs))
        for ys in range(Nys):
            y = sup2pix(ys, yscale, Npix=Ny, origin=0)
            sup_spec[:,ys,:] += np.nanmean(spec_supx[:,y[0]:y[-1]+1,:],axis=1)

        ## Do not extrapolate the wave grid that is not covered by input spectra
        ##-----------------------------------------------------------------------
        if extrapoff==True:
            for phot in filt:
                # w_grid = read_ascii(croot+'/lib/filt_'+phot, dtype=float)[:,0]
                # w_grid = ascii.read(croot+'/lib/filt_'+phot+ascext,
                #                     names=['Wave','Spectral Response'])['Wave']
                w_grid = read_hdf5(croot+'/lib/data/filt_'+phot,
                                   'Filter wavelength (microns)')
                # print(w_spec[0], w_grid[0])
                # print(w_spec[-1], w_grid[-1])
                if w_spec[0]>w_grid[0] or w_spec[-1]<w_grid[-1]:
                    warnings.warn('Synthetic photometry of {} can be underestimated' \
                                  'due to uncovered wavelengths'.format(phot))
            ## Insert 2 wvl (0.01 um & w_spec[0]-0.01 um) with 0 value
            wave = np.insert(w_spec, 0, (.01, w_spec[0]-.01))
            flux = np.insert(sup_spec, 0, np.zeros(sup_spec.shape[-1]), axis=0)
            flux = np.insert(flux, 0, np.zeros(sup_spec.shape[-1]), axis=0)
        else:
            wave = w_spec
            flux = sup_spec

        ## Write input.h5
        ##----------------
        fortIN = os.getcwd()+'/synthetic_photometry_input'
        
        write_hdf5(fortIN, 'Filter label', filt)
        write_hdf5(fortIN, 'Wavelength (microns)', wave, append=True)
        write_hdf5(fortIN, 'Flux (x.Hz-1)', flux, append=True)
        write_hdf5(fortIN, '(docalib,dophot)', [1,1], append=True)

        ## Call the Fortran lib
        ##----------------------
        SP.call('synthetic_photometry', shell=True)

        ## Read output.h5
        ##----------------
        fortOUT = os.getcwd()+'/synthetic_photometry_output'

        ds.wcen = read_hdf5(fortOUT, 'Central wavelength (microns)')
        ds.sup_filt = read_hdf5(fortOUT, 'Flux (x.Hz-1)')
        ds.smat = read_hdf5(fortOUT, 'Standard deviation matrix')

        ## Original pixels
        ds.Fnu_filt = np.zeros((Nw,Ny,Nx))
        for x in range(Nx):
            for y in range(Ny):
                xs = pix2sup(x, xscale, origin=0)
                ys = pix2sup(y, yscale, origin=0)
                ds.Fnu_filt[:,y,x] = ds.sup_filt[:,ys,xs]

        ## Convert zeros to NaNs
        ds.sup_filt[ds.sup_filt==0] = np.nan
        ds.Fnu_filt[ds.Fnu_filt==0] = np.nan

        ## Reform outputs
        if Ndim==1:
            ds.Fnu_filt = ds.Fnu_filt[:,0,0]
            ds.sup_filt = ds.sup_filt[:,0,0]
        if len(ds.wcen)==1:
            ds.wcen = ds.wcen[0]
            ds.Fnu_filt = ds.Fnu_filt[0]
            ds.sup_filt = ds.sup_filt[0]
            ds.smat = ds.smat[0][0]
        
        ## Clean temperary h5 files
        ##--------------------------
        if not verbose:
            SP.call('rm -rf '+fortIN+h5ext, shell=True, cwd=os.getcwd())
            SP.call('rm -rf '+fortOUT+h5ext, shell=True, cwd=os.getcwd())

        return ds

    def correct_spec(self, gain=1., offset=0., w_spec=None, Fnu_spec=None,
                     wlim=(None,None), ylim=(None,None), xlim=(None,None),
                     header=None, filOUT=None):
        '''
        Correct spectra
        
        ------ INPUT ------
        gain                scalar or ndarray (Default: 1.)
        offset              scalar or ndarray (Default: 0.)
        w_spec              wavelengths (Default: None - via filIN)
        Fnu_spec            spectra (Default: None - via filIN)
        wlim                wave limits (Default: (None,None))
        ylim                y limits (Default: (None,None))
        xlim                x limits (Default: (None,None))
        filOUT              overwrite fits file (Default: None)
        ------ OUTPUT ------
        new_spec            new_spec = gain * Fnu_spec + offset
        '''
        ## Input is a FITS file
        if self.filIN is not None:
            w_spec = self.wvl
            Fnu_spec = self.im

        w_spec = np.array(w_spec)
        Fnu_spec = np.array(Fnu_spec)
        if len(Fnu_spec.shape)==1:
            Ndim = 1
            Fnu_spec = Fnu_spec[:,np.newaxis,np.newaxis]
        else:
            Ndim = 3
        Nw, Ny, Nx = Fnu_spec.shape

        if np.isscalar(gain):
            a = np.array(gain)
            a = np.array(a[np.newaxis,np.newaxis])
            a = np.repeat(a[:,:], Ny, axis=0)
            a = np.repeat(a[:,:], Nx, axis=1)
        else:
            a = gain
        if np.isscalar(offset):
            b = np.array(offset)
            b = np.array(b[np.newaxis,np.newaxis])
            b = np.repeat(b[:,:], Ny, axis=0)
            b = np.repeat(b[:,:], Nx, axis=1)
        else:
            b = offset

        ## Truncate wavelengths
        if wlim[0] is None:
            wmin = w_spec[0]
        else:
            wmin = wlim[0]
        if wlim[1] is None:
            wmax = w_spec[-1]
        else:
            wmax = wlim[1]

        ## Crop map if 3D
        xmin, xmax = xlim
        ymin, ymax = ylim

        ## Modify spectra
        new_spec = np.copy(Fnu_spec)
        for k, lam in enumerate(w_spec):
            if lam>=wmin and lam<=wmax:
                new_spec[k,ymin:ymax,xmin:xmax] = \
                    a[ymin:ymax,xmin:xmax] * Fnu_spec[k,ymin:ymax,xmin:xmax] \
                    + b[ymin:ymax,xmin:xmax]

        ## Reform outputs
        if Ndim==1:
            new_spec = new_spec[:,0,0]
                    
        if filOUT is not None:
            if header is None:
                header = self.hdr
            write_fits(filOUT, header, new_spec, wave=w_spec)
        
        return new_spec

"""
class spec2phot(intercalib):
    '''
    Intercalibration between spectrometry and photometry (REF)

    --- INPUT ---
    filIN       to convolve
    filREF      convolution ref
    phot        photometry name (once a phot)
    filKER      convolution kernel(s) (Default: None)
    --- OUTPUT ---
    '''
    def __init__(self, filIN, filREF, phot, filKER=None, saveKER=None, \
        uncIN=None, Nmc=0, filOUT=None):
        super().__init__(filIN)
        self.phot = phot

        if self.wave is not None: # filIN is spec
            ## Convolve filIN (spec)
            if filKER is not None:
                conv = iconvolve(filIN, filKER, saveKER, \
                    uncIN, filOUT=filPRO)
            else:
                filPRO = filIN # filPRO is spec
            
            ## Reprojection to spec (filIN)
            pro = imontage(filREF, filPRO)
            F_phot = pro.reproject(filOUT=filOUT)

        else: # filIN is phot
            ## Reset header (should be spec)
            self.hdr = read_fits(filREF).header
            self.im = read_fits(filREF).data
            self.wvl = read_fits(filREF).wave
            
            ## Convolve filIN (phot)
            if filKER is not None:
                conv = iconvolve(filIN, filKER, saveKER, \
                    uncIN, filOUT=filPRO)
            else:
                filPRO = filIN # filPRO is phot
            
            ## Reprojection to spec (filREF)
            pro = imontage(filPRO, filREF)
            F_phot = pro.reproject(filOUT=filOUT)

        ## Synthetic photometry
        wcen, Fsyn, Fsig = self.synthetic_photometry((phot))
        self.wcen = wcen[0]
        self.Fsyn = Fsyn[0]
        self.Fsig = Fsig[0][0]

        self.gain = F_phot / self.Fsyn

    def calib_gain(self):
        return self.gain

    def image(self):
        return self.Fsyn
    
    def write_image(self, filSYN):
        comment = "Synthetic photometry with " + self.phot
        write_fits(filSYN, self.hdr, self.Fsyn, self.wvl, COMMENT=comment)

class phot2phot(intercalib):
    '''
    Intercalibration between two photometry
    '''
    def __init__(self, filIN, filREF, filKER=None, saveKER=None, \
        uncIN=None, Nmc=0, filOUT=None):

        ## Convolution (optional)
        if filKER is not None:
            conv = iconvolve(filIN, filKER, saveKER, \
                uncIN, filOUT=filPRO)
        else:
            filPRO = filIN

        ## Reprojection config
        pro = imontage(filPRO, filREF)
        self.im = pro.reproject(filOUT=filOUT)

    def image(self):
        return self.im
"""

def photometry_profile(datdir=None, *photometry):
    '''
    ------ INPUT ------
    datdir              profile data path (Default: ./lib/)
    photometry          photometry
    ------ OUTPUT ------
    '''
    ## Read data
    ##-----------
    lam = []
    val = []
    for phot in photometry:
        if datdir is None:
            # datdir = croot+'/lib/'
            datdir = croot+'/lib/data/'
        # dat = read_ascii(datdir+'filt_'+phot, dtype=float)
        # lam.append(dat[:,0])
        # val.append(dat[:,1])

        # dat = ascii.read(datdir+'filt_'+phot+ascext,
        #                  names=['Wave','Spectral Response'])
        # lam.append(dat['Wave'])
        # val.append(dat['Spectral Response'])

        lam.append(read_hdf5(datdir+'filt_'+phot,
            'Filter wavelength (microns)'))
        val.append(read_hdf5(datdir+'filt_'+phot,
            'Filter transmission'))
    # lam = np.array(lam)
    # val = np.array(val)

    ## Plotting setting
    ##------------------
    p = pplot(xlim=(1.9, 40.), ylim=(.01, 1.01),
              xlog=1, ylog=0,
              xlabel=r'$Wavelength,\,\,\lambda\,\,[\mu m]$',
              ylabel='Response',
              # ylabel='Spectral response\n(electrons / photon)',
              legend='upper left', legendalpha=.1,
              figsize=(12,3), title=None, clib='tableau')
    for i,w in enumerate(lam):
        p.add_plot(w, val[i], lw=1.8, label=photometry[i])

    p.set_fig(left=.05, bottom=.2, right=.99, top=.99)

    # sizeXL = 50
    # p.set_font(xticksize=sizeXL, yticksize=sizeXL, \
    #     axesize=sizeXL, legendsize=sizeXL)

    ## vlines (e.g. band markers)
    ##----------------------------
    greylines = []
    pinklines = []
    greylines.extend([2.3567863, 5.1532226]) # AKARI/IRC
    greylines.extend([5.242817, 7.597705]) # Spitzer/IRS-SL2
    pinklines.extend([7.3675313, 8.66892]) # Spitzer/IRS-SL3
    greylines.extend([7.5337057, 14.736635]) # Spitzer/IRS-SL1
    greylines.extend([14.266611, 21.051888]) # Spitzer/IRS-LL2
    pinklines.extend([19.483675, 21.50092]) # Spitzer/IRS-LL3
    greylines.extend([20.555237, 38.41488]) # Spitzer/IRS-LL1
    for gl in greylines:
        p.ax.axvline(gl, linestyle='dotted', color='grey')
    for pl in pinklines:
        p.ax.axvline(pl, linestyle='dotted', color='pink')

    ## tick setting
    ##-------------------- x --------------------------
    xtic = [2, 3, 4, 5, 6, 7, 8, 10, 12, 15, 20, 30, 40]
    xtic_min = np.arange(2., 41., 1.)
    p.ax.set_xticks(xtic, minor=False) # major
    p.ax.set_xticks(xtic_min, minor=True) # minor
    # ScalarFormatter().set_scientific(False)
    p.ax.xaxis.set_major_formatter(ScalarFormatter()) # major
    p.ax.xaxis.set_minor_formatter(NullFormatter()) # minor
    # p.ax.minorticks_off()
    ##--------------------- y --------------------------
    ytic = np.arange(0, 1.01, .2)
    ytic_min = np.arange(0, 1., .1)
    p.ax.set_yticks(ytic, minor=False) # major
    p.ax.set_yticks(ytic_min, minor=True) # minor
    # ScalarFormatter().set_scientific(False)
    p.ax.yaxis.set_major_formatter(ScalarFormatter()) # major
    p.ax.yaxis.set_minor_formatter(NullFormatter()) # minor
    # p.ax.minorticks_off()
    ##-----------------------------------------------

    return p

"""
------------------------------ MAIN (test) ------------------------------
"""
if __name__ == "__main__":

    pass



from setuptools import setup

setup(
    name = 'rapyuta',
    version = '2.1',
    author = 'D. HU',
    author_email = 'dangning.hu@outlook.com',
    description = 'libraRy of Astronomical PYthon UTilities for Astrophysics nerds',
    license = 'BSD',
    keywords = 'astronomy astrophysics astrometry imaging spectroscopy spitzer akari jwst',
    url = 'https://github.com/kxxdhdn/RAPYUTA',
    project_urls={
        'IDL': 'https://github.com/kxxdhdn/RAPYUTA/tree/main/idl',
        'SwING': 'https://github.com/kxxdhdn/RAPYUTA/tree/main/swing',
        'Tests': 'https://github.com/kxxdhdn/RAPYUTA/tree/main/tests',
    },

    python_requires='>=3.6',
    install_requires = [
        'numpy', 'scipy', 'matplotlib', 
        'astropy', 'reproject>=0.7.1', 'h5py', 'tqdm', 
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
    ],
    
    ## Plugins
    entry_points={
        # Installation test with command line
        'console_scripts': [
            'rapyutest = rapyuta:iTest',
        ],
    },

    ## Packages
    packages = ['rapyuta'],

    ## Package data
    package_data = {
        # include files in rapyuta/lib
        'rapyuta': ['lib/*.txt','lib/data/*.h5'],
    },
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""

Plots

    plotrange, Btau, Ctau, ellipse, SUE
    plotool:
        set_clib, set_fig, set_ax, 
        reset_handles, append_handles, get_handles, set_legend, 
        plot, eplot, save, show, close
    pplot(plotool):
        add_plot, add_legend

"""

import warnings
from astropy import units as u
import numpy as np
from scipy import optimize
# import matplotlib as mpl
from matplotlib.ticker import (
    NullFormatter, ScalarFormatter, LogFormatter,
    LogFormatterExponent, LogFormatterSciNotation,
    PercentFormatter
)
import matplotlib.colors as mplc
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

## Local
from utilities import InputError, merge_aliases
from arrays import arrayize, ramp

# cmap = mpl.cm.viridis
# norm = mpl.colors.Normalize(vmin=0, vmax=1)


##------------------------------
## Automatic plot range setting
##------------------------------
def plotrange(x,y,xran=None,yran=None,xlog=False,ylog=False,mask=None, \
              errx=None,erry=None,xisln=False,yisln=False):
    '''
    Automatically sets the x and y ranges for (X,Y) plots, based on the entered
    data set.

    Copyright: F. Galliano
    '''

    # Check the imput
    N = np.size(x)
    if (np.size(y) != N):
        UT.strike('plotrange','x and y should have the same size.')
    xran = arrayize(xran,N=2)
    yran = arrayize(yran,N=2)

    # X error bar settings
    if errx is not None:
        if (np.isscalar(errx)): errx = np.array([errx])
        sex = np.shape(errx)
        if (len(sex) == 2):
            if (sex != (2,N) ): UT.strike('plotrange','wrong size for errx.')
        elif (len(sex) == 1):
            if (sex != (N,) ): UT.strike('plotrange','wrong size for errx.')
            errx = np.array([errx,errx])
    else:
        errx = np.zeros((2,N))
        
    # Y error bar settings
    if erry is not None:
        if (np.isscalar(erry)): erry = np.array([erry])
        sey = np.shape(erry)
        if (len(sey) == 2):
            if (sey != (2,N) ): UT.strike('plotrange','wrong size for erry.')
        elif (len(sey) == 1):
            if (sey != (N,) ): UT.strike('plotrange','wrong size for erry.')
            erry = np.array([erry,erry])
    else:
        erry = np.zeros((2,N))
    
    # Homogenize the arrays and account for errors
    xlow = np.array(x,dtype=float).flatten() - errx[0,:]
    xhigh = xlow + errx[1,:]
    ylow = np.array(y,dtype=float).flatten() - erry[0,:]
    yhigh = ylow + erry[1,:]

    # Lin/Log
    if (xisln): xlow, xhigh = np.exp(xlow), np.exp(xhigh)
    if (yisln): ylow, yhigh = np.exp(ylow), np.exp(yhigh)
    
    # Mask
    mask = arrayize(mask,default=True,N=N)
    if (xlog): mask = ( mask & (xlow > 0) & (xhigh > 0) )
    if (ylog): mask = ( mask & (ylow > 0) & (yhigh > 0) )
    if (xran[0] != None): mask = ( mask & (xlow >= xran[0]) )
    if (xran[1] != None): mask = ( mask & (xhigh <= xran[1]) )
    if (yran[0] != None): mask = ( mask & (ylow >= yran[0]) )
    if (yran[1] != None): mask = ( mask & (yhigh <= yran[1]) )
            
    # Plain range
    xran = np.array([ np.min(xlow[mask]), np.max(xhigh[mask]) ])
    yran = np.array([ np.min(ylow[mask]), np.max(yhigh[mask]) ])

    # Add aesthetical margins
    fracmarg = 0.03
    if (not xlog):
        dxr = xran[1] - xran[0]
        xran += ( dxr*(-fracmarg), dxr*fracmarg )        
    else:
        dxr = xran[1] / xran[0]
        xran *= ( dxr**(-fracmarg), dxr**fracmarg )
    if (not ylog):
        dyr = yran[1] - yran[0]
        yran += ( dyr*(-fracmarg), dyr*fracmarg )
    else:
        dyr = yran[1] / yran[0]
        yran *= ( dyr**(-fracmarg), dyr**fracmarg )

    # Output
    return(xran,yran)

##-----------------------------------------------
##
##    Plotting functions for ellipses and SUEs
##
##           (Copyright: F. Galliano)
##
##-----------------------------------------------

## Function for SUEs
def Btau(tau):
    return( (1-2/np.pi)*(tau-1)**2 + tau )

## Function for SUEs
def Ctau(tau):
    return( ( (4/np.pi-1)*tau**2 + (3-8/np.pi)*tau + 4/np.pi-1 ) * (tau-1) )

## Ellipse
def ellipse(xmean=None,ymean=None,xstdev=None,ystdev=None,rho=None, \
            xmin=None,xmax=None,ymin=None,ymax=None,Npt=300, \
            xisln=False,yisln=False):
    """
    UNCERTAINTY ELLIPSES

    Function to plot uncertainty ellipses (or 1 sigma contour of a bivariate
    normal distribution). The parameters are the means (xmean,ymean), the 
    standard deviations (xstdev,ystdev) and the correlation coefficients (rho).
    The optional bounds (xmin,xmax,ymin,ymax) have the effect of truncating the
    ellipses in case there is a range of parameter space that is forbidden.

    It is important to notice that the xisln/yisln parameters are not related to
    the log settings of the axes where we plot the ellipse, but are here to 
    indicate that the moments of the variable to plot correspond to the natural 
    logarithm (ln) of the variable we want to display. For instance, for 
    displaying the ellipses of (x,y) where, for x, the moments are those of lnx,
    we would write:
        ellipse(xmean=mean_of_lnx,ymean=mean_of_y,xstdev=stdev_of_lnx, \
                ystdev=stdev_of_y,rho=correl_coeff_of_lnx_and_y,xisln=True)
    """
    x = ramp(x0=xmean-xstdev*(1-1.E-5),x1=xmean+xstdev*(1-1.E-5),N=Npt)
    c1 = rho * (x-xmean)/xstdev
    c2 = np.sqrt( (1-rho**2) * (1-(x-xmean)**2/xstdev**2) )
    y1 = ystdev * ( c1 - c2 ) + ymean
    y2 = ystdev * ( c1 + c2 ) + ymean
    xplot = np.concatenate((x,x[::-1],[x[0]]))
    yplot = np.concatenate((y1,y2[::-1],[y1[0]]))
    if (xisln): xplot = np.exp(xplot)
    if (yisln): yplot = np.exp(yplot)
    if (xmin != None): xplot[xplot < xmin] = xmin
    if (xmax != None): xplot[xplot > xmax] = xmax
    if (ymin != None): yplot[yplot < ymin] = ymin
    if (ymax != None): yplot[yplot > ymax] = ymax
    return(xplot,yplot)


## SUEs (1 sigma contour of a bivariate split-normal distribution)
def SUE(xmean=None,ymean=None,xstdev=None,ystdev=None,rho=None, \
        xskew=None,yskew=None,xmin=None,xmax=None,ymin=None,ymax=None, \
        Npt=300,xisln=False,yisln=False):
    """
    SKEWED UNCERTAINTY ELLIPSES (SUE)

    Function to plot uncertainty SUEs (or 1 sigma contour of a bivariate
    split-normal distribution). The parameters are the means (xmean,ymean), the 
    standard deviations (xstdev,ystdev), the skewnesses (xskew,yskew) and the 
    correlation coefficients (rho). The optional bounds (xmin,xmax,ymin,ymax) 
    have the effect of truncating the SUEs in case there is a range of 
    parameter space that is forbidden.

    It is important to notice that the xisln/yisln parameters are not related to
    the log settings of the axes where we plot the SUE, but are here to 
    indicate that the moments of the variable to plot correspond to the natural 
    logarithm (ln) of the variable we want to display. For instance, for 
    displaying the ellipses of (x,y) where, for x, the moments are those of lnx,
    we would write:
        SUE(xmean=mean_of_lnx,ymean=mean_of_y,xstdev=stdev_of_lnx, \
            ystdev=stdev_of_y,xskew=skewness_of_lnx,yskew=skewness_of_y, \
            rho=correl_coeff_of_lnx_and_y,xisln=True)
    """

    # Rotation angle
    theta = 1./2 * np.arctan( 2*rho*xstdev*ystdev / (xstdev**2-ystdev**2) )

    # Numerically solve for taux and tauy (tau=1.D2 ==> skew=0.99)
    taugrid = ramp(N=10000,x0=1.E-2,x1=1.E2,log=True)
    Ax = np.sqrt(np.pi/2) \
       * ( (np.cos(theta))**3*xskew*xstdev**3 \
         + (np.sin(theta))**3*yskew*ystdev**3 ) \
       / ( (np.sin(theta))**6 + (np.cos(theta))**6 ) \
       * ( ( (np.cos(theta))**2 - (np.sin(theta))**2 ) \
         / ( (np.cos(theta))**2*xstdev**2 \
           - (np.sin(theta))**2*ystdev**2 ) )**1.5
    Ay = np.sqrt(np.pi/2) \
       * ( (np.cos(theta))**3*yskew*ystdev**3 \
         - (np.sin(theta))**3*xskew*xstdev**3 ) \
       / ( (np.cos(theta))**6 + (np.sin(theta))**6 ) \
       * ( ( (np.cos(theta))**2 - (np.sin(theta))**2 ) \
         / ( (np.cos(theta))**2*ystdev**2 \
           - (np.sin(theta))**2*xstdev**2 ) )**1.5
    taux = np.exp(np.interp(Ax,Ctau(taugrid)/(Btau(taugrid))**1.5, \
                            np.log(taugrid)))
    tauy = np.exp(np.interp(Ay,Ctau(taugrid)/(Btau(taugrid))**1.5, \
                            np.log(taugrid)))
    if (not np.isfinite(taux) or taux > 1.E2): taux = 1.E2
    if (not np.isfinite(tauy) or tauy > 1.E2): tauy = 1.E2
 
    # Rest of the parameters
    lambdax = np.sqrt( ( (np.cos(theta))**2*xstdev**2 \
                       - (np.sin(theta))**2*ystdev**2 ) \
                     / ( (np.cos(theta))**2 - (np.sin(theta))**2 ) / Btau(taux) )
    lambday = np.sqrt( ( (np.cos(theta))**2*ystdev**2 \
                       - (np.sin(theta))**2*xstdev**2 ) \
                     / ( (np.cos(theta))**2 - (np.sin(theta))**2 ) / Btau(tauy) )
    x0 = xmean - np.sqrt(2/np.pi) * ( np.cos(theta)*lambdax*(taux-1) \
                                    - np.sin(theta)*lambday*(tauy-1) )
    y0 = ymean - np.sqrt(2/np.pi) * ( np.sin(theta)*lambdax*(taux-1) \
                                    + np.cos(theta)*lambday*(tauy-1) )

    # Draw the SUE
    matrot = np.array([ [ np.cos(theta), -np.sin(theta) ], \
                        [ np.sin(theta), np.cos(theta) ] ])
    xell_ax1 = np.zeros(2)
    yell_ax1 = np.zeros(2)
    xell_ax2 = np.zeros(2)
    yell_ax2 = np.zeros(2)
    for k in np.arange(4):
        if (k == 0):
            xell_sub = ramp(N=Npt,x0=-lambdax,x1=0) + x0
            rx = 1-(xell_sub-x0)**2/lambdax**2
            yell_sub = np.zeros(Npt)
            yell_sub[rx >= 0] = -lambday * np.sqrt(rx[rx >= 0]) + y0
            yell_sub[rx < 0] = np.nan
        elif (k == 1):
            xell_sub = ramp(N=Npt,x0=0,x1=lambdax*taux) + x0
            rx = 1-(xell_sub-x0)**2/lambdax**2/taux**2
            yell_sub = np.zeros(Npt)
            yell_sub[rx >= 0] = -lambday * np.sqrt(rx[rx >= 0]) + y0
            yell_sub[rx < 0] = np.nan
        elif (k == 2):
            xell_sub = (ramp(N=Npt,x0=0,x1=lambdax*taux))[::-1] + x0
            rx = 1-(xell_sub-x0)**2/lambdax**2/taux**2
            yell_sub = np.zeros(Npt)
            yell_sub[rx >= 0] = lambday*tauy * np.sqrt(rx[rx >= 0]) + y0
            yell_sub[rx < 0] = np.nan
        elif (k == 3):
            xell_sub = (ramp(N=Npt,x0=-lambdax,x1=0))[::-1] + x0
            rx = 1-(xell_sub-x0)**2/lambdax**2
            yell_sub = np.zeros(Npt)
            yell_sub[rx >= 0] = lambday*tauy * np.sqrt(rx[rx >= 0]) + y0
            yell_sub[rx < 0] = np.nan

        # Add the limit case (half ellipse)
        mask = np.logical_and(np.isfinite(yell_sub),np.isfinite(xell_sub))
        xell_sub = xell_sub[mask]
        yell_sub = yell_sub[mask]
        Nsub = np.count_nonzero(mask)

        # Rotate the ellipse
        for j in np.arange(Nsub):
            vecell = np.matmul(matrot, \
                               np.array([xell_sub[j]-x0,yell_sub[j]-y0]))
            xell_sub[j] = vecell[0] + x0
            yell_sub[j] = vecell[1] + y0
        if (k == 0):
            xell = xell_sub
            yell = yell_sub
        else:
            xell = np.concatenate((xell,xell_sub))
            yell = np.concatenate((yell,yell_sub))
    xplot = np.concatenate((xell,[xell[0]]))
    yplot = np.concatenate((yell,[yell[0]]))
                  
    # Logs and limits
    if (xisln):
        xplot = np.exp(xplot)
        x0 = np.exp(x0)
    if (yisln):
        yplot = np.exp(yplot)
        y0 = np.exp(y0)
    if (xmin != None):
        xplot[xplot < xmin] = xmin
        if (x0 < xmin): x0 = xmin
    if (xmax != None):
        xplot[xplot > xmax] = xmax
        if (x0 > xmax): x0 = xmax
    if (ymin != None):
        yplot[yplot < ymin] = ymin
        if (y0 < ymin): y0 = ymin
    if (ymax != None):
        yplot[yplot > ymax] = ymax
        if (y0 > ymax): y0 = ymax
    return(xplot,yplot,x0,y0)

##-----------------------------------------------
##
##            <plotool> based tools
##
##-----------------------------------------------

class plotool:
    '''
    plot Tool

    ------ INPUT ------
    nrows,ncols         Default: 1,1
    figint              Interactive figure (Default: True)

    plt.subplots(**kwargs)
    '''
    def __init__(self, nrows=1, ncols=1,
                 x=np.zeros(2), y=np.zeros(2), figint=True, **kwargs):
        '''
        ------ self ------
        figid               ID number of current figure (Default: 0)
        horder              ID number of current handle in self.handles (Default: 0)
        handles             Current handle list (Default: [])
        labels              Current label list (Default: [])
        legend              Current legend (Default: None)
        trans               transformation of coordinate system
                              'Data'
                              'Axes' (default)
                              'Figure'
                              'Display'
        '''
        ## INPUTS
        self.figid = 0
        self.horder = 0
        self.handles = []
        self.labels = []
        self.legend = None
        self.trans = 'Axes'
        self.x = x
        self.y = y

        if figint:
            plt.ion()

        self.nrows = nrows
        self.ncols = ncols

        self.fig, self.axes = plt.subplots(nrows, ncols, squeeze=False,
                                           num=self.figid, **kwargs)
        ## De-squeeze/Inflate (if squeeze=True)
        # if nrows==1 and ncols==1:
        #     self.axes = np.array(self.axes)[np.newaxis,np.newaxis]
        # elif nrows==1:
        #     self.axes = self.axes[np.newaxis,:]
        # elif ncols==1:
        #     self.axes = self.axes[:,np.newaxis]

        self.ax = self.axes[0,0]

    def set_clib(self, clib):
        if clib=='base':
            self.clib = list(mplc.BASE_COLORS) # 8 colors
        elif clib=='tableau':
            self.clib = list(mplc.TABLEAU_COLORS) # 10 colors
        elif clib=='ccs4' or clib=='x11':
            self.clib = list(mplc.CSS4_COLORS)
        elif clib=='xkcd':
            self.clib = list(mplc.XKCD_COLORS)
        else:
            self.clib = clib
        
    def set_fig(self, left=None, right=None, bottom=None, top=None,
                wspace=None, hspace=None, title=None, tsize=None):

        self.fig.subplots_adjust(left=left, right=right,
            bottom=bottom, top=top, wspace=wspace, hspace=hspace)
        
        if title is not None:
            self.fig.suptitle(title, size=tsize)
        
    def set_ax(self, subpos=None, # ax = axes[subpos[0],subpos[1]]
               xlog=False, ylog=False, # ax.set_xscale
               basex=10, basey=10, nonposx='clip', nonposy='clip', # ax.set_xscale
               xlim=(None,None), ylim=(None,None), #ax.set_xlim
               xtk=None, xtkmi=None, xtkform=None, xtksize=None, # ax.xaxis.set_tick_params
               ytk=None, ytkmi=None, ytkform=None, ytksize=None, 
               xlabel=None, xsize=None, ylabel=None, ysize=None, # ax.set_xlabel
               xtktoggle=False, ytktoggle=False,
               title=None, tsize=None, # ax.set_title (subplot title)
               **kwargs):
        '''
        nonposx, nonposy: 'sym', 'mask', 'clip'

        ------ INPUT ------
        subpos              identify subplot (Default: None - current self.ax)
        xtk,ytk             list of major tick labels
        xtkmi,ytkmi         list of minor tick labels
        xtkform,ytkform     formatter of (major) tick labels
        xtksize,ytksize     size of (major) tick labels
        xsize,ysize         size of axes labels
        xtktoggle           toggle x tick and label to top
        ytktoggle           toggle y tick and label to right

        self.ax.tick_params(**kwargs):
        axis                {'x', 'y', 'both' (default)}
        reset               Default: False
        which               {'major' (dafault), 'minor', 'both'}
        direction           {'in', 'out' (default), 'inout'}
        length              tick length
        width               tick width
        '''
        if self.nrows!=1 or self.ncols!=1:
            if subpos is not None:
                self.ax = self.axes[subpos[0],subpos[1]]
            ## else: keep current self.ax (Default: (0,0))

        if xlog:
            if nonposx=='sym':
                self.ax.set_xscale('symlog',base=basex)
            else:
                self.ax.set_xscale('log',base=basex,nonpositive=nonposx)
        if ylog:
            if nonposx=='sym':
                self.ax.set_yscale('symlog',base=basey)
            else:
                self.ax.set_yscale('log',base=basey,nonpositive=nonposy)

        if xlim[0]!=None or xlim[1]!=None:
            self.ax.set_xlim(xlim[0], xlim[1])
        if ylim[0]!=None or ylim[1]!=None:
            self.ax.set_ylim(ylim[0], ylim[1])

        ## x ticks
        if xtkform=='log':
            xformat = LogFormatter()
        elif xtkform=='log_exp':
            xformat = LogFormatterExponent()
        elif xtkform=='log_sci':
            xformat = LogFormatterSciNotation()
        elif xtkform=='pct':
            xformat = PercentFormatter()
        else:
            xformat = ScalarFormatter()
        if xtk is not None:
            self.ax.set_xticks(xtk, minor=False) # major
        if xtkmi is not None:
            self.ax.set_xticks(xtkmi, minor=True) # minor
        self.ax.xaxis.set_major_formatter(xformat) # major
        self.ax.xaxis.set_minor_formatter(NullFormatter()) # minor
        self.ax.xaxis.set_tick_params(labelsize=xtksize)
        if xtktoggle:
            self.ax.xaxis.tick_top()
            self.ax.xaxis.set_label_position('top')
        ## y ticks
        if ytkform=='log':
            yformat = LogFormatter()
        elif ytkform=='log_exp':
            yformat = LogFormatterExponent()
        elif ytkform=='log_sci':
            yformat = LogFormatterSciNotation()
        elif ytkform=='pct':
            yformat = PercentFormatter()
        else:
            yformat = ScalarFormatter()
        if ytk is not None:
            self.ax.set_yticks(ytk, minor=False) # major
        if ytkmi is not None:
            self.ax.set_yticks(ytkmi, minor=True) # minor
        self.ax.yaxis.set_major_formatter(yformat) # major
        self.ax.yaxis.set_minor_formatter(NullFormatter()) # minor
        self.ax.yaxis.set_tick_params(labelsize=ytksize)
        if ytktoggle:
            self.ax.yaxis.tick_right()
            self.ax.yaxis.set_label_position('right')
        ## both ticks
        self.ax.tick_params(**kwargs)
        
        if xlabel is not None:
            self.ax.set_xlabel(xlabel, size=xsize)
        if ylabel is not None:
            self.ax.set_ylabel(ylabel, size=ysize)
        if title is not None:
            self.ax.set_title(title, size=tsize)

    def reset_handles(self):
        '''
        Reset handles (before plot with new legend)
        '''
        if self.trans=='Figure':
            self.fig.add_artist(self.legend)
        elif self.trans=='Axes' or self.trans=='Data':
            self.ax.add_artist(self.legend)
        self.handles = []
        self.horder = 0

        return self.handles

    def append_handles(self):
        '''
        Append currently added handle (after plot)
        '''
        handles, labels = self.ax.get_legend_handles_labels()
        for handle, label in zip(handles, labels):
            if label not in self.labels:
                self.handles.append(handle)
                self.labels.append(label)
                self.horder += 1

        return self.handles
        
    def get_handles(self):
        '''
        Get non-repeated handles (after plot)
        '''
        handles, labels = self.ax.get_legend_handles_labels()
        handlist, labelist = [], []
        for handle, label in zip(handles, labels):
            if label not in labelist:
                handlist.append(handle)
                labelist.append(label)
        self.horder = len(handlist)
        self.handles = handlist
        self.labels = labelist

        return self.handles
        
    def set_legend(self, subpos=None,
                   left=None, right=None, bottom=None, top=None, figtight=False,
                   handles=None, **kwargs):
        '''
        - bbox_to_anchor rules: (1,1) correspond to upper right of the axis

                    bbox_to_anchor = (1,1)
          .--------.
          |        |
          |  Axes  |
          | (bbox) |
          |        |
          .--------.
        
        - lengend loc is relative to the reference point (bbox_to_anchor) as follows:

                               |
          lower right (2)      |       lower left (1)
          ---------------bbox_to_anchor---------------
          upper right (3)      |       upper left (4)
                               |
        
        ------ INPUT ------
        subpos              identify subplot (Default: None - current self.ax)
        x0,y0               bottom left of current Axes (Default: None)
        shrinkx,shrinky     current Axes size (Default: 1)
        figtight            ignore Axes settings (Default: False)
        handles             Default: None - self.handles

        self.fig/ax.legend(**kwargs):
        title               title of legend box
        loc                 relative position of legend box
        bbox_to_anchor      reference point of legend box
        bbox_transform      reference frame of coordinates
                              self.fig.transFigure
                              self.ax.transAxes (default)
        ------ OUTPUT ------
        self.legend         registrated with self.fig/ax.add_artist(self.legend)
        '''
        if handles is None:
            handles = self.handles

        if self.trans=='Figure':
            self.fig.subplots_adjust(left=left, right=right,
                                     bottom=bottom, top=top)
            
            self.legend = self.fig.legend(handles=handles, **kwargs)
        elif self.trans=='Axes' or self.trans=='Data':
            if self.nrows!=1 or self.ncols!=1:
                if subpos is not None:
                    self.ax = self.axes[subpos[0],subpos[1]]
                ## else: keep current self.ax (Default: (0,0))
            
            ## Set current axes
            bbox = self.ax.get_position()
            if left is None:
                x0 = bbox.x0
            else:
                x0 = bbox.x0 + left*(bbox.x1-bbox.x0)
            if right is None:
                x1 = bbox.x1
            else:
                x1 = bbox.x0 + right*(bbox.x1-bbox.x0)
            if bottom is None:
                y0 = bbox.y0
            else:
                y0 = bbox.y0 + bottom*(bbox.y1-bbox.y0)
            if top is None:
                y1 = bbox.y1
            else:
                y1 = bbox.y0 + top*(bbox.y1-bbox.y0)
            self.ax.set_position([x0, y0, x1-x0, y1-y0])

            self.legend = self.ax.legend(handles=handles, **kwargs)
        else:
            raise InputError('<plotool.set_legend>',
                             'Non-recognized transformation!')

        if figtight:
            self.fig.tight_layout()

        return self.legend
        
    def plot(self, x=None, y=None, yerr=None, xerr=None, xisln=False, yisln=False,
             fmt='', capsize=None, barsabove=False, # errorbar kw
             ecolor=None, ec=None, elinewidth=None, elw=None, # errorbar kw
             subpos=None, mod='CA', **kwargs):
        '''
        Like set_ax(), this is a clump operation.
        The idea is to all set in one command,
        while each single operation should also be valid.

        ------ INPUT ------
        subpos              identify subplot (Default: None - current self.ax)

        self.ax.errorbar(**kwargs)
        '''
        if self.nrows!=1 or self.ncols!=1:
            if subpos is not None:
                self.ax = self.axes[subpos[0],subpos[1]]
            ## else: keep current self.ax (Default: (0,0))

        ## Keyword aliases
        ec = merge_aliases(None, ecolor=ecolor, ec=ec)
        elw = merge_aliases(None, elinewidth=elinewidth, elw=elw)
        
        if x is None:
            x = self.x
        else:
            self.x = x
        if y is None:
            y = self.y
        else:
            self.y = y

        ## Log inputs
        xp, xperr, yp, yperr = None, None, None, None
        if (xisln):
            if x is not None: xp = np.exp(x)
            if xerr is not None: xperr = x * (1. - np.exp(-xerr)) ## suppose xmin <-> lnxmin
        else:
            if x is not None: xp = x
            if xerr is not None: xperr = xerr
        if (yisln):
            if y is not None: yp = np.exp(y)
            if yerr is not None: yperr = y * (1. - np.exp(-yerr))
        else:
            if y is not None: yp = y
            if yerr is not None: yperr = yerr

        ## CA: Cartesian using matplotlib.pyplot.errorbar
        if mod=='CA':
                
            self.markers, self.caps, self.bars = self.ax.errorbar(
                x=xp, y=yp, yerr=yperr, xerr=xperr,
                fmt=fmt, ecolor=ec, elinewidth=elw,
                capsize=capsize, barsabove=barsabove,
                **kwargs)
            
        else:
            print('*******************')
            print('Prochainement...')
            
            print('PL: polar')
            print('CL: cylindrical')
            print('SP: spherical')
            print('*******************')

    def eplot(self, x=None, y=None, mask=None,
              xmin=None, xmax=None, ymin=None, ymax=None,
              xisln=False, yisln=False,
              ## Uncertainty kw
              sigmax=None, sigmay=None, rho=None,
              gammax=None, gammay=None,
              ## Bar/ellipse keywords
              edgecolor=None, ecolor=None, ec=None,
              elinewidth=None, elw=None, elinestyle=None, els=None,
              efillcolor=None, efc=None, efill=False, ehatch=None,
              errinlegend=None, alpha=1,
              ## Other kw
              subpos=None, **kwargs):
        '''
        DISPLAY ERROR BARS/ELLIPSES/SUES
        
        x and y are (N,) shape arrays. sigmax and sigmay are either (N,) or (2,N)
        shape arrays. If xisln is set, then it is assumed that all the x related
        quantities are moments of lnx (x is mu(lnx), sigmax is stdev(lnx), etc.).
        This is independent of the xlog setting.

        If rho is an (N,) shape array, an ellipse
        is drawn, instead. If gammax or gammay are not None, an "asymmetric 
        ellipse" or SUE (1 sigma contour of a bivariate split-normal 
        distribution) is drawn. 

        ------ INPUT ------
        subpos              identify subplot (Default: None - current self.ax)

        self.plot(**kwargs)
        '''
        if self.nrows!=1 or self.ncols!=1:
            if subpos is not None:
                self.ax = self.axes[subpos[0],subpos[1]]
            ## else: keep current self.ax (Default: (0,0))
            
        ## kw aliases
        ec = merge_aliases(None, edgecolor=edgecolor, ecolor=ecolor, ec=ec)
        elw = merge_aliases(None, elinewidth=elinewidth, elw=elw)
        els = merge_aliases(None, elinestyle=elinestyle, els=els)
        efc = merge_aliases(None, efillcolor=efillcolor, efc=efc)
        
        ## Central values
        xp, N = arrayize(x)
        yp = arrayize(y,N=N)
        if (N != np.size(yp)):
            raise InputError('<plotool.eplot>',
                             'x and y must have the same size.')
            
        ## Ellipse/SUE
        ell = (rho is not None)
        skewll = ( ( (gammax is not None) or (gammay is not None) ) \
                   and ell )

        ## X error bar settings
        uncx = False
        if sigmax is not None:
            uncx = True
            if (np.isscalar(sigmax)): sigmax = np.array([sigmax])
            sex = np.shape(sigmax)
            if (len(sex) == 2):
                if (sex != (2,N) ):
                    raise InputError('<plotool.eplot>',
                                     'wrong size for errx.')
            elif (len(sex) == 1):
                if (sex != (N,) ):
                    raise InputError('<plotool.eplot>',
                                     'wrong size for errx.')
                sigmax = np.array([sigmax,sigmax])

        ## Y error bar settings
        uncy = False
        if sigmay is not None:
            uncy = True
            if (np.isscalar(sigmay)): sigmay = np.array([sigmay])
            sey = np.shape(sigmay)
            if (len(sey) == 2):
                if (sey != (2,N) ):
                    raise InputError('<plotool.eplot>',
                                     'wrong size for erry.')
            elif (len(sey) == 1):
                if (sey != (N,) ):
                    raise InputError('<plotool.eplot>',
                                     'wrong size for erry.')
                sigmay = np.array([sigmay,sigmay])
        
        if (ell):
            ## Setting the error bars and the potential lower/upper limits
            if (not uncx or not uncy):
                raise InputError('<plotool.eplot>',
                                 'for ellipses, both x and y errors must be set.')
            xcenell = xp.copy()
            xerr = sigmax[0,:]
            if (xisln):
                xcen = np.exp(xp)
            else:
                xcen = xp.copy()
            ycenell = yp.copy()
            yerr = sigmay[0,:]
            if (yisln):
                ycen = np.exp(yp)
            else:
                ycen = yp.copy()
                
            ## Mask
            if mask is None:
                mask = np.ones(N,bool)
            Nsub = np.count_nonzero(mask)
            xcen = xcen[mask]
            ycen = ycen[mask]
            xcenell = xcenell[mask]
            ycenell = ycenell[mask]
            xerr = xerr[mask]
            yerr = yerr[mask]
            rho = arrayize(rho,N=N)
            rho = rho[mask]
            if (skewll):
                gammax = arrayize(gammax,N=N)
                gammay = arrayize(gammay,N=N)
                gammax = gammax[mask]
                gammay = gammay[mask]

            ## Lines
            elw = arrayize(elw,N=N)
            elw = elw[mask]
            els = arrayize(els,N=N)
            els = els[mask]

            ## Colors
            efill = arrayize(efill,N=N)
            if ec is None:
                ec = np.array(['b' for i in range(N)])
                efc = np.array(['lightblue' for i in range(N)])
            else:
                ec = arrayize(ec,N=N)
                efc = arrayize(efc,N=N)
            ec = ec[mask]
            efc = efc[mask]

            ## Hatching the ellipses
            if (ehatch):
                ehatch = np.array(['back slanted' for i in np.arange(N)])
            elif (not ehatch):
                ehatch = np.array(['' for i in np.arange(N)])
            else:
                ehatch = arrayize(ehatch,default='',N=N)
            ehatch = ehatch[mask]
            nothatched = (ehatch == '')
            efill = np.logical_or(efill,(ehatch != ''))
            
            ## Draw the error bars
            if (not skewll):

                for i in np.arange(Nsub):
                    if i==0:
                        elegend = errinlegend
                    else:
                        elegend = None
                    xell, yell = ellipse(xmean=xcenell[i],ymean=ycenell[i],
                                         xstdev=xerr[i],ystdev=yerr[i],
                                         rho=rho[i],xmin=xmin,xmax=xmax,
                                         ymin=ymin,ymax=ymax,xisln=xisln,yisln=yisln)
                    self.plot(xell, yell,
                              color=ec[i], linewidth=elw[i], linestyle=els[i],
                              label=elegend, alpha=alpha, **kwargs)
                    if (efill[i]):
                        self.ax.fill(xell,yell,hatch=ehatch[i],fill=nothatched[i],
                                     color=efc[i],alpha=alpha)

            elif(skewll):

                for i in np.arange(Nsub):
                    if i==0:
                        elegend = errinlegend
                    else:
                        elegend = None
                    xell, yell, x0, y0 \
                        = SUE(xmean=xcenell[i],ymean=ycenell[i],
                              xstdev=xerr[i],ystdev=yerr[i],
                              xskew=gammax[i],yskew=gammay[i],
                              rho=rho[i],xmin=xmin,xmax=xmax,
                              ymin=ymin,ymax=ymax,
                              xisln=xisln,yisln=yisln)
                    xcen[i] = x0
                    ycen[i] = y0
                    self.plot(xell, yell,
                              color=ec[i], linewidth=elw[i], linestyle=els[i],
                              label=elegend, alpha=alpha, **kwargs)
                    if (efill[i]):
                        self.ax.fill(xell,yell,hatch=ehatch[i],fill=nothatched[i],
                                     color=efc[i],alpha=alpha)
            
    def save(self, savename=None, transparent=False,
             figtight=False, close=True, **kwargs):
        '''

        ------ INPUT ------
        self.fig.savefig(**kwargs)
        '''
        ## Use self.fig.set_size_inches to define ranges
        if figtight:
            bbox = 'tight'
        else:
            bbox = None
            
        if savename is not None:
            self.fig.savefig(savename, transparent=transparent,
                             bbox_inches=bbox, **kwargs)
        else:
            warnings.warn('Not saved! ')

        if (close): plt.close(self.figid)

    def show(self):

        plt.ioff()
        self.fig.show(self.fig)

    def close(self):

        plt.close(self.figid)

class pplot(plotool):
    '''
    Uni-frame plot (1 row * 1 col)

    ------ INPUT ------
    plotool.plot(**kwargs)
    '''
    def __init__(self, x=None, y=None, yerr=None, xerr=None,
                 figsize=(8,6), figint=False,
                 ## errorbar kw
                 fmt='', capsize=None, barsabove=False,
                 ## eplot kw
                 edgecolor='grey', ecolor='grey', ec='grey',
                 elinewidth=None, elw=None, elinestyle=None, els=None,
                 mask=None, xmin=None, xmax=None, ymin=None, ymax=None,
                 sigmax=None, sigmay=None, rho=None,
                 gammax=None, gammay=None,xisln=False, yisln=False,
                 efillcolor=None, efc=None, efill=None, ehatch=None,
                 errinlegend=None, alpha=1,
                 ## set_fig kw
                 left=.15, bottom=.1, right=.95, top=.9,
                 wspace=None, hspace=None, title=None, titlesize=20,
                 ## set_ax kw
                 xlog=None, ylog=None,
                 basex=10, basey=10, nonposx='clip', nonposy='clip',
                 xlim=(None, None), ylim=(None,None),
                 tk=None, tkmi=None, tkform=None, tksize=None, # same xy
                 xtk=None, xtkmi=None, xtkform=None, xtksize=15, # x
                 ytk=None, ytkmi=None, ytkform=None, ytksize=15, # y
                 xlabel='X', ylabel='Y', xsize=None, ysize=None, xysize=15,
                 ## set_legend kw
                 loc=None, legendsize=15, legendalpha=1,
                 anchor=None, figtight=False,
                 ## Other kw
                 clib='base', c=None, **kwargs):
        super().__init__(figsize=figsize, figint=figint, x=x, y=y)

        self.pltid = 0

        ## Keyword aliases
        ## Note that ec and ecolor are shared by plot (errorbar) and eplot (ellipse)
        ec = merge_aliases('grey', edgecolor=edgecolor, ecolor=ecolor, ec=ec) # Default: 'grey'
        elw = merge_aliases(None, elinewidth=elinewidth, elw=elw)
        els = merge_aliases(None, elinestyle=elinestyle, els=els)
        efc = merge_aliases(None, efillcolor=efillcolor, efc=efc)

        ## Auto color
        self.set_clib(clib)
        if c is None:
            c = self.clib[self.pltid]

        ## set_fig
        self.set_fig(left=left, bottom=bottom, right=right, top=top,
            wspace=wspace, hspace=hspace, title=title, tsize=titlesize)

        ## set_ax
        if tk is not None:
            xtk = tk
            ytk = tk
        if tkmi is not None:
            xtkmi = tkmi
            ytkmi = tkmi
        if tkform is not None:
            xtkform = tkform
            ytkform = tkform
        if tksize is not None:
            xtksize = tksize
            ytksize = tksize
        if xysize is not None:
            xsize = xysize
            ysize = xysize
        self.set_ax(xlog=xlog, ylog=ylog,
                    basex=basex, basey=basey, nonposx=nonposx, nonposy=nonposy,
                    xlim=xlim, ylim=ylim,
                    xtk=xtk, xtkmi=xtkmi, xtkform=xtkform, xtksize=xtksize,
                    ytk=ytk, ytkmi=ytkmi, ytkform=ytkform, ytksize=ytksize,
                    xlabel=xlabel, xsize=xsize, ylabel=ylabel, ysize=ysize)

        ## plot
        ell = (rho is not None)
        if (not ell):
            
            self.plot(x=x, y=y, yerr=yerr, xerr=xerr,
                      xisln=xisln, yisln=yisln,
                      fmt=fmt, ec=ec, elw=elw, els=els, # errorbar kw
                      capsize=capsize, barsabove=barsabove, # errorbar kw
                      c=c, alpha=alpha, **kwargs)
            self.append_handles()
            
        else:

            self.plot(x=x, y=y, xisln=xisln, yisln=yisln,
                      fmt=fmt, c=c, alpha=alpha, **kwargs)
            self.append_handles()
            ## Ellipse errors
            self.eplot(x=x, y=y, mask=mask,
                       xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                       xisln=xisln, yisln=yisln,
                       sigmax=sigmax, sigmay=sigmay, rho=rho,
                       gammax=gammax, gammay=gammay,
                       ec=ec, elw=elw, els=els,
                       efill=efill, efc=efc, ehatch=ehatch,
                       errinlegend=errinlegend, alpha=alpha)
            self.append_handles()
            ## Replace line by patch in legend
            epatch = mpatches.Patch(ec=ec, lw=elw, ls=elw,
                                    fill=efill, fc=efc, alpha=alpha,
                                    hatch=ehatch, label=errinlegend)
            self.handles[-1] = epatch

        ## set_legend
        if loc is not None:
            self.set_legend(loc=loc, fontsize=legendsize,
                            bbox_to_anchor=anchor, figtight=figtight,
                            framealpha=legendalpha)
        self.loc = loc
        self.legendsize = legendsize
        self.anchor = anchor
        self.figtight = figtight
        self.legendalpha = legendalpha

    def add_plot(self, x=None, y=None, yerr=None, xerr=None,
                 ## errorbar kw
                 fmt='', capsize=None, barsabove=False,
                 ## eplot kw
                 edgecolor='grey', ecolor='grey', ec='grey',
                 elinewidth=None, elw=None, elinestyle=None, els=None,
                 mask=None, xmin=None, xmax=None, ymin=None, ymax=None,
                 sigmax=None, sigmay=None, rho=None,
                 gammax=None, gammay=None,xisln=False, yisln=False,
                 efillcolor=None, efc=None, efill=None, ehatch=None,
                 errinlegend=None, alpha=1,
                 ## set_legend kw (only when addlegend=True)
                 addlegend=False, loc=None, legendsize=15,
                 legendalpha=1, anchor=None, figtight=False,
                 ## Other (errorbar) kw
                 c=None, label=None, **kwargs):
        '''
        ------ INPUT ------
        plotool.plot(**kwargs)
        '''
        self.pltid += 1

        ## Keyword aliases
        ## Note that ec and ecolor are shared by plot (errorbar) and eplot (ellipse)
        ec = merge_aliases('grey', edgecolor=edgecolor, ecolor=ecolor, ec=ec) # Default: 'grey'
        elw = merge_aliases(None, elinewidth=elinewidth, elw=elw)
        els = merge_aliases(None, elinestyle=elinestyle, els=els)
        efc = merge_aliases(None, efillcolor=efillcolor, efc=efc)

        ## Auto color
        if self.pltid==len(self.clib):
            self.pltid = 0
        if c is None:
            c = self.clib[self.pltid]
        
        if x is None:
            x = self.x
        else:
            self.x = x
        if y is None:
            y = self.y
        else:
            self.y = y

        ## plot
        if addlegend:
            self.reset_handles()
            self.loc = loc
            self.legendsize = legendsize
            self.anchor = anchor
            self.figtight = figtight
            self.legendalpha = legendalpha
            
        ell = (rho is not None)
        if (not ell):
            
            self.plot(x=x, y=y, yerr=yerr, xerr=xerr,
                      xisln=xisln, yisln=yisln,
                      fmt=fmt, ec=ec, elw=elw, els=els, # errorbar kw
                      capsize=capsize, barsabove=barsabove, # errorbar kw
                      c=c, alpha=alpha, label=label, **kwargs)
            self.append_handles()
            
        else:

            self.plot(x=x, y=y, xisln=xisln, yisln=yisln,
                      fmt=fmt, c=c, alpha=alpha,
                      label=label, **kwargs)
            self.append_handles()
            ## Ellipse errors
            self.eplot(x=x, y=y, mask=mask,
                       xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                       xisln=xisln, yisln=yisln,
                       sigmax=sigmax, sigmay=sigmay, rho=rho,
                       gammax=gammax, gammay=gammay,
                       ec=ec, elw=elw, els=els,
                       efill=efill, efc=efc, ehatch=ehatch,
                       errinlegend=errinlegend, alpha=alpha)
            self.append_handles()
            ## Replace line by patch in legend
            epatch = mpatches.Patch(ec=ec, lw=elw, ls=els,
                                    fill=efill, fc=efc, alpha=alpha,
                                    hatch=ehatch, label=errinlegend)
            self.handles[-1] = epatch

        ## set_legend
        if (self.loc is not None) and (label is not None):
            self.set_legend(loc=self.loc, fontsize=self.legendsize,
                            bbox_to_anchor=self.anchor, figtight=self.figtight,
                            framealpha=self.legendalpha)
